import java.util.ArrayList;
import java.util.LinkedList;

import aiinterface.AIInterface;
import aiinterface.CommandCenter;
import enumerate.Action;
import enumerate.State;
import simulator.Simulator;
import struct.CharacterData;
import struct.FrameData;
import struct.GameData;
import struct.Key;
import struct.MotionData;

public class Poring implements AIInterface {

  private Simulator simulator;
  private Key key;
  private CommandCenter commandCenter;
  private boolean playerNumber;
  private GameData gameData;

  /** Main FrameData */
  private FrameData frameData;

  /** Data with FRAME_AHEAD frames ahead of FrameData */
  private FrameData simulatorAheadFrameData;

  /** All actions that could be performed by self character */
  private LinkedList<Action> myActions;

  /** All actions that could be performed by the opponent character */
  private LinkedList<Action> oppActions;

  /** self information */
  private CharacterData myCharacter;

  /** opponent information */
  private CharacterData oppCharacter;

  /** Number of adjusted frames (following the same recipe in JerryMizunoAI) */
  private static final int FRAME_AHEAD = 14;

  private ArrayList<MotionData> myMotion;

  private ArrayList<MotionData> oppMotion;

  private Action[] actionAir;

  private Action[] LudList;

  private Action[] Nearby;

  private Action[] Escape;

  private Action [] LudFight;

  private Action[] actionGround;

  private Action[] GarnetFight;

  private Action[] goFace;

  private Action spSkill;

  private Node rootNode;

  /** True if in debug mode, which will output related log */
  public static final boolean DEBUG_MODE = false;

  @Override
  public void close() {}

	@Override
	public void roundEnd(int p1Hp, int p2Hp, int frames){

	}

  @Override
  public void getInformation(FrameData frameData) {
    this.frameData = frameData;
    this.commandCenter.setFrameData(this.frameData, playerNumber);

    myCharacter = frameData.getCharacter(playerNumber);
    oppCharacter = frameData.getCharacter(!playerNumber);
  }

  @Override
  public int initialize(GameData gameData, boolean playerNumber) {
    this.playerNumber = playerNumber;
    this.gameData = gameData;
    this.key = new Key();
    this.frameData = new FrameData();
    this.commandCenter = new CommandCenter();

    this.myActions = new LinkedList<Action>();
    this.oppActions = new LinkedList<Action>();

    simulator = gameData.getSimulator();

    actionAir =
        new Action[] {Action.AIR_GUARD, Action.AIR_A, Action.AIR_B, Action.AIR_DA, Action.AIR_DB,
            Action.AIR_FA, Action.AIR_FB, Action.AIR_UA, Action.AIR_UB, Action.AIR_D_DF_FA,
            Action.AIR_D_DF_FB, Action.AIR_F_D_DFA, Action.AIR_F_D_DFB, Action.AIR_D_DB_BA,
            Action.AIR_D_DB_BB};

    actionGround =
        new Action[] {Action.STAND_D_DB_BA, Action.FORWARD_WALK, Action.DASH,
            Action.JUMP, Action.FOR_JUMP, Action.BACK_JUMP, Action.STAND_GUARD,
            Action.THROW_B, Action.STAND_B,
            Action.CROUCH_A, Action.CROUCH_B, Action.STAND_FA, Action.STAND_FB, Action.CROUCH_FA,
            Action.CROUCH_FB, Action.STAND_D_DF_FA, Action.STAND_D_DF_FB, Action.STAND_F_D_DFA,
            Action.STAND_F_D_DFB, Action.STAND_D_DB_BB};

    spSkill = Action.STAND_D_DF_FC;

    GarnetFight =
    	new Action[] {Action.STAND_D_DB_BA, /*Action.FORWARD_WALK, Action.DASH,*/
            Action.JUMP, Action.FOR_JUMP, Action.BACK_JUMP, Action.STAND_GUARD,
            Action.THROW_B, Action.STAND_B,
            Action.STAND_FA, Action.STAND_FB, Action.CROUCH_FA,
            Action.CROUCH_FB, Action.STAND_D_DF_FA, /*Action.STAND_D_DF_FB,*/ Action.STAND_F_D_DFA,
            Action.STAND_F_D_DFB, Action.STAND_D_DB_BB};

    goFace =
    		new Action[] {
    				Action.FORWARD_WALK,Action.FOR_JUMP,
    				Action.DASH, Action.BACK_JUMP
    		};

    LudFight =
    		new Action[] {Action.STAND_D_DB_BA, Action.FORWARD_WALK, Action.DASH,
            Action.JUMP, Action.BACK_JUMP, Action.STAND_GUARD,
            Action.STAND_B,
            Action.CROUCH_A, Action.CROUCH_B, Action.STAND_FB, /*Action.CROUCH_FA,*/
            Action.CROUCH_FB, Action.STAND_D_DF_FA, Action.STAND_D_DF_FB, Action.STAND_F_D_DFA,
            Action.STAND_D_DB_BB, Action.STAND_F_D_DFB};

    LudList	=
    		new Action[] {Action.CROUCH_FB
    				};
    Nearby =
    		new Action[] {Action.FORWARD_WALK};

    Escape =
    		new Action[] {Action.FOR_JUMP, Action.STAND_GUARD,
    				Action.CROUCH_GUARD, Action.STAND_FA, Action.STAND_F_D_DFB,
    				Action.CROUCH_FB, Action.THROW_B};

    myMotion = gameData.getMotionData(playerNumber);
    oppMotion = gameData.getMotionData(!playerNumber);

    return 0;
  }

  @Override
  public Key input() {
    return key;
  }

  @Override
  public void processing() {
    if (canProcessing()) {
      if (commandCenter.getSkillFlag()) {
        key = commandCenter.getSkillKey();
      } else { // play
		  key.empty();
		  commandCenter.skillCancel();
    	  if(frameData.getRemainingFramesNumber() > 0 && gameData.getCharacterName(playerNumber).equals("ZEN") && rulebased() == true)
    	  {
    		  ZenRun();
    	  }
		  else {// mcTs
			  mctsPrepare(); // Some preparation for MCTS
			  rootNode =
					  new Node(simulatorAheadFrameData, null, myActions, oppActions, gameData, playerNumber,
                commandCenter);
			  rootNode.createNode();
			  Action bestAction = rootNode.mcts(); // Perform MCTS
			  if (Poring.DEBUG_MODE)
			  {
				  rootNode.printNode(rootNode);
				  }
			  commandCenter.commandCall(bestAction.name()); // Perform an action selected by MCTS
			  }
	  }
    }
  }

  /**
   * Determine whether or not the AI can perform an action
   *
   * @return whether or not the AI can perform an action
   */
  public boolean canProcessing() {
    return !frameData.getEmptyFlag() && frameData.getRemainingTime() > 0;
  }

  /**
   * Some preparation for MCTS
   * Perform the process for obtaining FrameData with 14 frames ahead
   */
  public void mctsPrepare() {
    simulatorAheadFrameData = simulator.simulate(frameData, playerNumber, null, null, FRAME_AHEAD);

    myCharacter = simulatorAheadFrameData.getCharacter(playerNumber);
    oppCharacter = simulatorAheadFrameData.getCharacter(!playerNumber);

    setMyAction();
    setOppAction();
  }

  public void setMyAction() {
    myActions.clear();

    int energy = myCharacter.getEnergy();

    if (myCharacter.getState() == State.AIR) {
      for (int i = 0; i < actionAir.length; i++) {
        if (Math.abs(myMotion.get(Action.valueOf(actionAir[i].name()).ordinal())
            .getAttackStartAddEnergy()) <= energy) {
          myActions.add(actionAir[i]);
        }
      }
    }
    if(gameData.getCharacterName(playerNumber).equals("LUD")){
    	System.out.println("HP = : "+(frameData.getCharacter(playerNumber).getHp() - frameData.getCharacter(!playerNumber).getHp() <= 0));
    	if(frameData.getDistanceX() <= 600 && frameData.getDistanceX() >= 370 && (frameData.getCharacter(playerNumber).getHp() - frameData.getCharacter(!playerNumber).getHp() <= 0)){
    		for (int i = 0; i < Nearby.length; i++) {
    	        if (Math.abs(myMotion.get(Action.valueOf(Nearby[i].name()).ordinal())
    	            .getAttackStartAddEnergy()) <= energy) {
    	          myActions.add(Nearby[i]);
    	          System.out.println("Nearby");
    	        }
    	      }
    	}
    	else if(frameData.getCharacter(playerNumber).getX() >= 680 || frameData.getCharacter(playerNumber).getX() == -170) {
    		for (int i = 0; i < Escape.length; i++) {
    	        if (Math.abs(myMotion.get(Action.valueOf(Escape[i].name()).ordinal())
    	            .getAttackStartAddEnergy()) <= energy) {
    	          myActions.add(Escape[i]);
    	          System.out.println("Escpe");
    	        }
    	      }
    	}
    	else if(frameData.getCharacter(playerNumber).getX() >= 490 && frameData.getCharacter(playerNumber).getX() <= 620 && frameData.getDistanceX() <= 190 && frameData.getDistanceX() >= 75 ||
    			frameData.getCharacter(playerNumber).getX() <= 36 && frameData.getCharacter(playerNumber).getX() >= -170 && frameData.getDistanceX() <= 190 && frameData.getDistanceX() >= 75) {
    		for (int i = 0; i < LudList.length; i++) {
    	        if (Math.abs(myMotion.get(Action.valueOf(Escape[i].name()).ordinal())
    	            .getAttackStartAddEnergy()) <= energy) {
    	          myActions.add(LudList[i]);
    	          System.out.println("Lud Test");
    	        }
    	      }
    	        }
    	else {
    		for (int i = 0; i < LudFight.length; i++) {
    	        if (Math.abs(myMotion.get(Action.valueOf(LudFight[i].name()).ordinal())
    	            .getAttackStartAddEnergy()) <= energy) {
    	          myActions.add(LudFight[i]);
    	          System.out.println("LudFight");
    	          System.out.println(frameData.getDistanceX());
    	        }
    	      }
    	}
    }
    if(gameData.getCharacterName(playerNumber).equals("GARNET")){
    	if(frameData.getDistanceX() <= 600 && frameData.getDistanceX() >= 370 && (frameData.getCharacter(playerNumber).getHp() - frameData.getCharacter(!playerNumber).getHp() <= 0)){
    		for (int i = 0; i < goFace.length; i++) {
    	        if (Math.abs(myMotion.get(Action.valueOf(goFace[i].name()).ordinal())
    	            .getAttackStartAddEnergy()) <= energy) {
    	          myActions.add(goFace[i]);
    	          System.out.println("goFace");
    	        }
    	      }
    	}

    	else if(frameData.getCharacter(playerNumber).getX() >= 530 && frameData.getCharacter(playerNumber).getX() <= 620 && frameData.getDistanceX() <= 150 && frameData.getDistanceX() >= 50 ||
    			frameData.getCharacter(playerNumber).getX() <= -11 && frameData.getCharacter(playerNumber).getX() >= -170 && frameData.getDistanceX() <= 150 && frameData.getDistanceX() >= 50) {
    		for (int i = 0; i < LudList.length; i++) {
    	        if (Math.abs(myMotion.get(Action.valueOf(Escape[i].name()).ordinal())
    	            .getAttackStartAddEnergy()) <= energy) {
    	          myActions.add(LudList[i]);
    	          System.out.println("Lud Test");
    	        }
    	      }
    	        }
    	else if(frameData.getCharacter(!playerNumber).getAction().toString() == "STAND_D_DF_FB") {
    		for (int i = 0; i < LudList.length; i++) {
    	        if (Math.abs(myMotion.get(Action.valueOf(Escape[i].name()).ordinal())
    	            .getAttackStartAddEnergy()) <= energy) {
    	          myActions.add(LudList[i]);
    	          System.out.println("Garnet Flee");
    	        }
    	      }
		  }
    	else
    	{
    		for (int i = 0; i < GarnetFight.length; i++) {
    	        if (Math.abs(myMotion.get(Action.valueOf(GarnetFight[i].name()).ordinal())
    	            .getAttackStartAddEnergy()) <= energy) {
    	          myActions.add(GarnetFight[i]);
    	          System.out.println("Stage Out");
    	        }
    	        }
    		}
    	}
    }
  public boolean rulebased()
  {
	  System.out.println(frameData.getCharacter(playerNumber).getX());
	  if(frameData.getCharacter(!playerNumber).getEnergy() <= 300 || frameData.getCharacter(playerNumber).getAction().toString().equals("STAND_D_DF_FC"))
	  {
		  return true;
	  }
	  else
	  {
		  return false;
	  }
  }
  public void ZenRun()
  {
	  if((frameData.getRemainingTime() < 10000 && frameData.getCharacter(playerNumber).getEnergy()>=300) || (frameData.getRemainingTime() < 10000 && frameData.getCharacter(playerNumber).getEnergy() >= 50) )
		{
			if(frameData.getCharacter(playerNumber).getEnergy() >= 300)
			{
				commandCenter.commandCall("STAND_D_DF_FC");
			}
			if(frameData.getRemainingTime() < 2000 && frameData.getDistanceX() < 50)
			{
				commandCenter.commandCall("STAND_F_D_DFB");
			}
			if(frameData.getRemainingTime() <2000)
			{
				commandCenter.commandCall("STAND_D_DF_FB");
			}
		}
		if(frameData.getCharacter(!playerNumber).getState().equals(State.AIR) && frameData.getCharacter(playerNumber).getState().equals(State.AIR)){
			if(frameData.getCharacter(playerNumber).getEnergy() > 50 && frameData.getDistanceX() > 80)
				commandCenter.commandCall("AIR_D_DF_FB");
			else if(frameData.getCharacter(playerNumber).getEnergy() > 5 && frameData.getCharacter(!playerNumber).getY() > frameData.getCharacter(playerNumber).getY()){
				commandCenter.commandCall("AIR_DB");
			}
			else
				commandCenter.commandCall("B");
		}
		if(frameData.getCharacter(!playerNumber).getState().equals(State.AIR) && !frameData.getCharacter(playerNumber).getState().equals(State.DOWN) && !frameData.getCharacter(playerNumber).getState().equals(State.AIR)){
			if(frameData.getDistanceY() < 50)
				commandCenter.commandCall("STAND_F_D_DFA");
		}
		if(frameData.getCharacter(playerNumber).getState().equals(State.STAND)){
			if((frameData.getDistanceX() > 300) && (frameData.getDistanceX() < 550)){
				commandCenter.commandCall("STAND_D_DF_FA");
			}
			else if(frameData.getCharacter(playerNumber).getX() >= 680){
				commandCenter.commandCall("FOR_JUMP");
			}
			else if(frameData.getCharacter(playerNumber).getX() <= -120){
				commandCenter.commandCall("FOR_JUMP");
			}
			else
				commandCenter.commandCall("FOR_JUMP");
		}
		else
			commandCenter.commandCall("B");

		}
  public void setOppAction() {
    oppActions.clear();

    int energy = oppCharacter.getEnergy();

    if (oppCharacter.getState() == State.AIR) {
      for (int i = 0; i < actionAir.length; i++) {
        if (Math.abs(oppMotion.get(Action.valueOf(actionAir[i].name()).ordinal())
            .getAttackStartAddEnergy()) <= energy) {
          oppActions.add(actionAir[i]);
        }
      }
    } else {
      if (Math.abs(oppMotion.get(Action.valueOf(spSkill.name()).ordinal())
          .getAttackStartAddEnergy()) <= energy) {
        oppActions.add(spSkill);
      }

      for (int i = 0; i < actionGround.length; i++) {
        if (Math.abs(oppMotion.get(Action.valueOf(actionGround[i].name()).ordinal())
            .getAttackStartAddEnergy()) <= energy) {
          oppActions.add(actionGround[i]);
        }
      }
    }
  }
}