import java.util.ArrayList;
import java.util.LinkedList;

import aiinterface.AIInterface;
import aiinterface.CommandCenter;
import enumerate.Action;
import enumerate.State;
import simulator.Simulator;
import struct.CharacterData;
import struct.FrameData;
import struct.GameData;
import struct.Key;
import struct.MotionData;

/**
 * EMCTS: Extended MCTS(Incorporate Motion of Oppornent to MCT and Consider Energy)
 * ZEN: Machete plus EMCTS
 * GARNET: EMCTS
 * LUD:EMCTS (Consider Distance)
 *
 * @author Eita Aoki
 */
public class Thunder01 implements AIInterface {

  private Simulator simulator;
  private Key key;
  private CommandCenter commandCenter;
  private boolean playerNumber;
  private GameData gameData;

  /** 大本のFrameData */
  private FrameData frameData;

  /** 大本よりFRAME_AHEAD分遅れたFrameData */
  private FrameData simulatorAheadFrameData;

  /** 自分が行える行動全て */
  private LinkedList<Action> myActions;

  /** 相手が行える行動全て */
  private LinkedList<Action> oppActions;

  /** 自分の情報 */
  private CharacterData myCharacter;

  /** 相手の情報 */
  private CharacterData oppCharacter;

  /** フレームの調整用時間(JerryMizunoAIを参考) */
  private static final int FRAME_AHEAD = 14;

  private ArrayList<MotionData> myMotion;

  private ArrayList<MotionData> oppMotion;

  private Action[] actionAir;

  private Action[] actionGround;

  private Action spSkill;

  private ExtendedNode rootNode;

  /** デバッグモードであるかどうか。trueの場合、様々なログが出力される */
  public static final boolean DEBUG_MODE = false;

  private CharacterName charName;

  @Override
  public void close() {}

  @Override
  public void getInformation(FrameData frameData) {
    this.frameData = frameData;
    this.commandCenter.setFrameData(this.frameData, playerNumber);

    myCharacter = frameData.getCharacter(playerNumber);
    oppCharacter = frameData.getCharacter(!playerNumber);
  }

  @Override
  public int initialize(GameData gameData, boolean playerNumber) {

    this.playerNumber = playerNumber;
    this.gameData = gameData;

    this.key = new Key();
    this.frameData = new FrameData();
    this.commandCenter = new CommandCenter();

    this.myActions = new LinkedList<Action>();
    this.oppActions = new LinkedList<Action>();

    simulator = gameData.getSimulator();
    System.out.println("thunder:"+gameData.getCharacterName(this.playerNumber));
    actionAir =
        new Action[] {Action.AIR_GUARD, Action.AIR_A, Action.AIR_B, Action.AIR_DA, Action.AIR_DB,
            Action.AIR_FA, Action.AIR_FB, Action.AIR_UA, Action.AIR_UB, Action.AIR_D_DF_FA,
            Action.AIR_D_DF_FB, Action.AIR_F_D_DFA, Action.AIR_F_D_DFB, Action.AIR_D_DB_BA,
            Action.AIR_D_DB_BB};
    actionGround =
        new Action[] {Action.STAND_D_DB_BA, Action.BACK_STEP, Action.FORWARD_WALK, Action.DASH,
            Action.JUMP, Action.FOR_JUMP, Action.BACK_JUMP, Action.STAND_GUARD,
            Action.CROUCH_GUARD, Action.THROW_A, Action.THROW_B, Action.STAND_A, Action.STAND_B,
            Action.CROUCH_A, Action.CROUCH_B, Action.STAND_FA, Action.STAND_FB, Action.CROUCH_FA,
            Action.CROUCH_FB, Action.STAND_D_DF_FA, Action.STAND_D_DF_FB, Action.STAND_F_D_DFA,
            Action.STAND_F_D_DFB, Action.STAND_D_DB_BB};
    spSkill = Action.STAND_D_DF_FC;

    myMotion = gameData.getMotionData(playerNumber);
    oppMotion = gameData.getMotionData(!playerNumber);


    String tmpcharname=this.gameData.getCharacterName(this.playerNumber);
    if(tmpcharname.equals("ZEN"))charName=CharacterName.ZEN;
    else if(tmpcharname.equals("GARNET"))charName=CharacterName.GARNET;
    else if(tmpcharname.equals("LUD"))charName=CharacterName.LUD;
    else charName=CharacterName.OTHER;

    return 0;
  }

  @Override
  public Key input() {
    return key;
  }

  private void mctsProcessing(){
	  rootNode =
	            new ExtendedNode(charName,myCharacter.getEnergy(),oppCharacter.getEnergy(), simulatorAheadFrameData, null, myActions, oppActions, gameData, playerNumber,
	                commandCenter);
	        rootNode.createNode();

	        Action bestAction = rootNode.mcts(); // MCTSの実行


	        commandCenter.commandCall(bestAction.name()); // MCTSで選択された行動を実行する
  }
  private void zenProcessing(){
	  FrameData tmpFrameData = simulator.simulate(frameData, this.playerNumber, null, null, 17);
	  CommandCenter cc=this.commandCenter;
	  cc.setFrameData(tmpFrameData, playerNumber);
		int distance = tmpFrameData.getDistanceX();
		int energy = frameData.getCharacter(playerNumber).getEnergy();
		CharacterData my = tmpFrameData.getCharacter(playerNumber);
		CharacterData opp = tmpFrameData.getCharacter(!playerNumber);
		int xDifference = my.getX() - opp.getX();

	  if ((opp.getEnergy() >= 150) && ((my.getHp() - opp.getHp()) <= 120))
			cc.commandCall("FOR_JUMP _B B B");
			// if the opp has 300 of energy, it is dangerous, so better jump!!
			// if the health difference is high we are dominating so we are fearless :)
		else if (!my.getState().equals(State.AIR) && !my.getState().equals(State.DOWN)) { //if not in air
			if ((distance > 150)) {
				cc.commandCall("FOR_JUMP"); //If its too far, then jump to get closer fast
			}
			else if (energy >= 150)
				cc.commandCall("STAND_D_DF_FC"); //High energy projectile
			else if ((distance > 100) && (energy >= 50))
				cc.commandCall("STAND_D_DB_BB"); //Perform a slide kick
			else if (opp.getState().equals(State.AIR)) //if enemy on Air
				cc.commandCall("STAND_F_D_DFA"); //Perform a big punch
			else if (distance > 100)
				this.mctsProcessing();
				//cc.commandCall("6 6 6"); // Perform a quick dash to get closer
			else
				this.mctsProcessing();
				//cc.commandCall("B"); //Perform a kick in all other cases, introduces randomness
		}
		else if ((distance <= 150) && (my.getState().equals(State.AIR) || my.getState().equals(State.DOWN))
				&& (((gameData.getStageWidth() - my.getX())>=200) || (xDifference > 0))
				&& ((my.getX() >=200) || xDifference < 0)) { //Conditions to handle game corners
			if (energy >= 5)
				this.mctsProcessing();
				//cc.commandCall("AIR_DB"); // Perform air down kick when in air
			else
				this.mctsProcessing();
				//cc.commandCall("B"); //Perform a kick in all other cases, introduces randomness
		}
		else
			this.mctsProcessing();
			//cc.commandCall("B"); //Perform a kick in all other cases, introduces randomness
  }

  private void garnetProcessing(){
	  this.mctsProcessing();
  }


  private boolean printnameflag=true;
  @Override
  public void processing() {

	  if (canProcessing()) {
		  if (commandCenter.getSkillFlag()) {
			  key = commandCenter.getSkillKey();
		  } else {
			  key.empty();
			  commandCenter.skillCancel();

			  mctsPrepare(); // MCTSの下準備を行う

			  if(charName==CharacterName.ZEN){
				  zenProcessing();
				  if(printnameflag)System.out.println("zenProcessing");
			  }else if(charName==CharacterName.GARNET){
				  garnetProcessing();
				  if(printnameflag)System.out.println("garnetProcessing");
			  }
			  else{
				  if(printnameflag)System.out.println("elseProcessing");
				  mctsProcessing();
			  }
			  printnameflag=false;

		  }
	  }
  }

  @Override
  public void roundEnd(int p1Hp, int p2Hp, int frames) {
	  key.empty();
	  commandCenter.skillCancel();
  }
  /**
   * AIが行動できるかどうかを判別する
   *
   * @return AIが行動できるかどうか
   */
  public boolean canProcessing() {
    return !frameData.getEmptyFlag() && frameData.getRemainingTime() > 0;
  }

  /**
   * MCTSの下準備 <br>
   * 14フレーム進ませたFrameDataの取得などを行う
   */
  public void mctsPrepare() {
    simulatorAheadFrameData = simulator.simulate(frameData, playerNumber, null, null, FRAME_AHEAD);

    myCharacter = simulatorAheadFrameData.getCharacter(playerNumber);
    oppCharacter = simulatorAheadFrameData.getCharacter(!playerNumber);

    setMyAction();
    setOppAction();
  }

  public void setMyAction() {
    myActions.clear();

    int energy = myCharacter.getEnergy();

    if (myCharacter.getState() == State.AIR) {
      for (int i = 0; i < actionAir.length; i++) {
        if (Math.abs(myMotion.get(Action.valueOf(actionAir[i].name()).ordinal())
            .getAttackStartAddEnergy()) <= energy) {
          myActions.add(actionAir[i]);
        }
      }
    } else {
      if (Math.abs(myMotion.get(Action.valueOf(spSkill.name()).ordinal())
          .getAttackStartAddEnergy()) <= energy) {
        myActions.add(spSkill);
      }

      for (int i = 0; i < actionGround.length; i++) {
        if (Math.abs(myMotion.get(Action.valueOf(actionGround[i].name()).ordinal())
            .getAttackStartAddEnergy()) <= energy) {
          myActions.add(actionGround[i]);
        }
      }
    }

  }

  public void setOppAction() {
    oppActions.clear();

    int energy = oppCharacter.getEnergy();

    if (oppCharacter.getState() == State.AIR) {
      for (int i = 0; i < actionAir.length; i++) {
        if (Math.abs(oppMotion.get(Action.valueOf(actionAir[i].name()).ordinal())
            .getAttackStartAddEnergy()) <= energy) {
          oppActions.add(actionAir[i]);
        }
      }
    } else {
      if (Math.abs(oppMotion.get(Action.valueOf(spSkill.name()).ordinal())
          .getAttackStartAddEnergy()) <= energy) {
        oppActions.add(spSkill);
      }

      for (int i = 0; i < actionGround.length; i++) {
        if (Math.abs(oppMotion.get(Action.valueOf(actionGround[i].name()).ordinal())
            .getAttackStartAddEnergy()) <= energy) {
          oppActions.add(actionGround[i]);
        }
      }
    }
  }
}
