import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;

import aiinterface.AIInterface;
import aiinterface.CommandCenter;
import enumerate.Action;
import enumerate.State;
import simulator.Simulator;
import struct.AttackData;
import struct.CharacterData;
import struct.FrameData;
import struct.GameData;
import struct.Key;
import struct.MotionData;

public class Ranezi implements AIInterface {

	private enum JumpState {
		RISING, PEAK, FALLING, ON_GROUND
	};

	private Simulator simulator;
	private Key key;
	private CommandCenter commandCenter;
	private boolean playerNumber;
	private GameData gameData;

	private FrameData frameData;

	private FrameData simulatorAheadFrameData;

	private LinkedList<Action> myActions;

	private LinkedList<Action> oppActions;

	private LinkedList<Integer> oppLastYPos;

	private LinkedList<Action> oppLastMoves;

	private CharacterData myCharacter;

	private CharacterData oppCharacter;

	private static final int FRAME_AHEAD = 14;

	private ArrayList<MotionData> myMotion;

	private ArrayList<MotionData> oppMotion;

	private Action[] actionAir;

	private Action[] actionGround;

	private Action spSkill;

	private Node rootNode;

	public static final boolean DEBUG_MODE = false;

	@Override
	public void close() {
		System.out.println((playerNumber?"p1 ":"p2 ") + num);
	}

	@Override
	public void getInformation(FrameData frameData) {
		this.frameData = frameData;
		this.commandCenter.setFrameData(this.frameData, playerNumber);

		myCharacter = frameData.getCharacter(playerNumber);
		oppCharacter = frameData.getCharacter(!playerNumber);

	}

	@Override
	public int initialize(GameData gameData, boolean playerNumber) {
		this.playerNumber = playerNumber;
		this.gameData = gameData;

		this.key = new Key();
		this.frameData = new FrameData();
		this.commandCenter = new CommandCenter();

		this.myActions = new LinkedList<Action>();
		this.oppActions = new LinkedList<Action>();
		this.oppLastYPos = new LinkedList<Integer>();
		this.oppLastMoves = new LinkedList<Action>();

		simulator = gameData.getSimulator();

		actionAir = new Action[] { Action.AIR_GUARD, Action.AIR_A, Action.AIR_B, Action.AIR_DA, Action.AIR_DB,
				Action.AIR_FA, Action.AIR_FB, Action.AIR_UA, Action.AIR_UB, Action.AIR_F_D_DFA, Action.AIR_F_D_DFB,
				Action.AIR_D_DB_BA, Action.AIR_D_DB_BB, Action.AIR_D_DF_FB, Action.AIR_D_DF_FA };

		actionGround = new Action[] { Action.STAND_D_DB_BA, Action.BACK_STEP, Action.FORWARD_WALK, Action.DASH,
				Action.JUMP, Action.FOR_JUMP, Action.BACK_JUMP, Action.STAND_GUARD, Action.CROUCH_GUARD, Action.THROW_A,
				Action.THROW_B, Action.STAND_A, Action.STAND_B, Action.CROUCH_A, Action.CROUCH_B, Action.STAND_FA,
				Action.STAND_FB, Action.CROUCH_FA, Action.CROUCH_FB, Action.STAND_D_DF_FA, Action.STAND_D_DF_FB,
				Action.STAND_F_D_DFA, Action.STAND_F_D_DFB, Action.STAND_D_DB_BB };
		spSkill = Action.STAND_D_DF_FC;

		myMotion = gameData.getMotionData(playerNumber);
		oppMotion = gameData.getMotionData(!playerNumber);

		return 0;
	}

	@Override
	public Key input() {
		return key;
	}
	int num = 0;
	@Override
	public void processing() {

		if (canProcessing()) {
			num++;
			if (commandCenter.getSkillFlag()) {
				key = commandCenter.getSkillKey();
			} else {
				key.empty();
				commandCenter.skillCancel();

				mctsPrepare();
				rootNode = new Node(simulatorAheadFrameData, null, myActions, oppActions, gameData, playerNumber,
						commandCenter);
				rootNode.createNode();

				Action bestAction = rootNode.mcts();
				if (Ranezi.DEBUG_MODE) {
					rootNode.printNode(rootNode);
				}
				commandCenter.commandCall(bestAction.name());
			}
		}

	}

	public boolean canProcessing() {
		return !frameData.getEmptyFlag() && frameData.getRemainingTime() > 0;
	}

	public void mctsPrepare() {
		simulatorAheadFrameData = simulator.simulate(frameData, playerNumber, null, null, FRAME_AHEAD);

		myCharacter = simulatorAheadFrameData.getCharacter(playerNumber);
		oppCharacter = simulatorAheadFrameData.getCharacter(!playerNumber);
		oppLastYPos.add(oppCharacter.getY());
		if (oppLastYPos.size() > 3) {
			oppLastYPos.removeFirst();
		}

		oppLastMoves.add(oppCharacter.getAction());
		if (oppLastMoves.size() > 10)
			oppLastMoves.removeFirst();

		setMyAction();
		setOppAction();
	}

	/*
	 * Check whether the enemy has launched at least one projectile
	 */
	private boolean checkForProjectile() {
		boolean opponentProjectile = false;
		int ownNumber = (this.playerNumber) ? 0 : 1;
		Iterator<AttackData> itr = frameData.getProjectiles().iterator();
		while (itr.hasNext() && !opponentProjectile) {
			AttackData current = itr.next();
			if (current.getPlayerNumber() != ownNumber)
				opponentProjectile = true;
		}
		return opponentProjectile;
	}

	/*
	 * Check whether the enemy is at the start or the middle or the end of his
	 * jump process
	 */
	private JumpState oppFindJumpState() {
		if (oppLastYPos.size() < 3)
			return JumpState.ON_GROUND;
		int currY = oppLastYPos.get(2);
		int lastY = oppLastYPos.get(1);
		int last2Y = oppLastYPos.get(0);

		if (oppCharacter.getState() != State.AIR)
			return JumpState.ON_GROUND;

		if (lastY <= last2Y) {
			if (currY < lastY)
				return JumpState.RISING;
			else if (currY > lastY)
				return JumpState.PEAK;
		} else {
			return JumpState.FALLING;
		}

		return JumpState.ON_GROUND;
	}

	public void setMyAction() {
		myActions.clear();

		int distanceX = frameData.getDistanceX();
		boolean opponentProjectile = checkForProjectile();
		JumpState oppJumpState = oppFindJumpState();

		int energy = myCharacter.getEnergy();

		if (myCharacter.getState() == State.AIR) {
			for (int i = 0; i < actionAir.length; i++) {
				if (Math.abs(myMotion.get(Action.valueOf(actionAir[i].name()).ordinal())
						.getAttackStartAddEnergy()) <= energy) {
					myActions.add(actionAir[i]);
				}
			}
		} else {

			if (Math.abs(
					myMotion.get(Action.valueOf(spSkill.name()).ordinal()).getAttackStartAddEnergy()) <= energy) {
				myActions.add(spSkill);
			}
			// Check whether there is any projectile and they are not too close
			// to each other
			if (opponentProjectile && distanceX > gameData.getStageWidth() / 4) {
				myActions.add(Action.BACK_JUMP);
			}
			// Check whether there is any projectile and they are not too close
			// to each other
			else if (opponentProjectile && distanceX > gameData.getStageWidth() / 6) {

				myActions.add(Action.FOR_JUMP);
				// Check whether the players are far away to each other
			} else if (distanceX > gameData.getStageWidth() / 2) {
				myActions.add(Action.FORWARD_WALK);

			}
			// Check whether the enemy jumps up
			else if (oppJumpState == JumpState.RISING) {
				myActions.add(Action.STAND_FB);
				myActions.add(Action.STAND_F_D_DFA);

			}

			else {
				for (int i = 0; i < actionGround.length; i++) {
					if (Math.abs(myMotion.get(Action.valueOf(actionGround[i].name()).ordinal())
							.getAttackStartAddEnergy()) <= energy) {
						myActions.add(actionGround[i]);
					}
				}
			}
		}

	}

	public void setOppAction() {
		oppActions.clear();

		int energy = oppCharacter.getEnergy();

		if (oppCharacter.getState() == State.AIR) {
			for (int i = 0; i < actionAir.length; i++) {
				if (Math.abs(oppMotion.get(Action.valueOf(actionAir[i].name()).ordinal())
						.getAttackStartAddEnergy()) <= energy) {
					oppActions.add(actionAir[i]);
				}
			}
		} else {
			if (Math.abs(oppMotion.get(Action.valueOf(spSkill.name()).ordinal())
					.getAttackStartAddEnergy()) <= energy) {
				oppActions.add(spSkill);
			}

			for (int i = 0; i < actionGround.length; i++) {
				if (Math.abs(oppMotion.get(Action.valueOf(actionGround[i].name()).ordinal())
						.getAttackStartAddEnergy()) <= energy) {
					oppActions.add(actionGround[i]);
				}
			}
		}
	}

	@Override
	public void roundEnd(int arg0, int arg1, int arg2) {
		// TODO 自動生成されたメソッド・スタブ

	}
}
